import json
import boto3
import logging
import sagemaker
from datetime import datetime
from omegaconf import OmegaConf
from importlib import import_module
from sagemaker.model_monitor import DataCaptureConfig
from sagemaker import (
    Predictor,
    ModelPackage,
    get_execution_role,
    Session
)
from typing import Dict, Optional
from .helpers import (
    get_datetime_str,
    ensure_min_length,
    get_approved_package,
    load_json_from_s3,
    get_configs,
    _normalize_pipeline_name
)
from mlops_sm import helpers

logger = logging.getLogger(__name__)


def upsert_pipeline(
        pipeline_module: str,
        pipeline_package: str,
        pipeline_name: str,
        pipeline_tags: Optional[Dict[str, str]] = None,
        dryrun: bool = False,
        *args,
):
    """
    Performs Sagemaker pipeline creating or updating.

    Example:
    >>> upsert_pipeline('training_pipeline', 'src', 'a_cool_pipeline_name', 'role-arn', ...)

    First two arguments is set so to follow (folder) package structure described below
    when the function is invoked from the root dir:
        root
        |
        |-- src
        |   |-- training_pipeline.py
        |   |-- ...
        |   `-- ...
        |
        ...

    :param pipeline_module: a "module path" within the 'pipeline_package' (relative to the 'pipeline_package' root)
    :param pipeline_package: a package where 'pipeline_module' is defined
    :param pipeline_name: the name of the pipeline
    :param pipeline_tags: {"<key>": "<value>", ...} dict to be set as SM pipeline resource tags
    :param dryrun: whether to skip actual pipeline upsert or not
    :param args: extra configuration to pass to pipeline building;
        must follow dot-notation (https://omegaconf.readthedocs.io/en/2.0_branch/usage.html#from-a-dot-list)
    """
    mod_pipe = import_module(f'.{pipeline_module}', pipeline_package)
    result_conf = get_configs(mod_pipe, pipeline_module, args)

    if logger.isEnabledFor(logging.INFO):
        logger.info('Result config:\n%s', OmegaConf.to_yaml(result_conf, resolve=True))
    sm_session = Session(default_bucket=OmegaConf.select(result_conf, 'pipeline.default_bucket', default=None))

    pipeline_name = _normalize_pipeline_name(pipeline_name)
    pipe_def = mod_pipe.get_pipeline(sm_session, pipeline_name, result_conf)
    if logger.isEnabledFor(logging.INFO):
        logger.info("Pipeline definition:\n%s",
                    json.dumps(
                        json.loads(pipe_def.definition()),
                        indent=2
                    ))

    if not dryrun:
        if pipeline_tags is not None:
            pipeline_tags = helpers.convert_param_dict_to_key_value_list(pipeline_tags)
        pipe_def.upsert(result_conf.pipeline.role, tags=pipeline_tags)


def run_pipeline(
        pipeline_name,
        execution_name_prefix,
        dryrun=False,
        **pipe_params):
    sm = boto3.client('sagemaker')
    now = datetime.today()
    now_str = get_datetime_str(now)
    pipe_exec_name = f'{execution_name_prefix}-{now_str}'
    start_pipe_args = {
        'PipelineName': pipeline_name,
        'PipelineExecutionDisplayName': pipe_exec_name,
        'PipelineParameters': [
            {
                'Name': k,
                'Value': str(v)
            }
            for k, v in pipe_params.items()
        ],
        'ClientRequestToken': ensure_min_length(pipe_exec_name, 32)
    }
    if dryrun:
        return start_pipe_args
    else:
        return sm.start_pipeline_execution(**start_pipe_args)


def deploy_model(model_package_group_name, instance_type, instance_count, endpoint_name, data_capture_s3_uri):
    instance_count = int(instance_count)

    boto_sess = boto3.Session()
    sm = boto_sess.client("sagemaker")
    sagemaker_session = sagemaker.Session(boto_session=boto_sess)

    pck = get_approved_package(sm, model_package_group_name)
    model_description = sm.describe_model_package(ModelPackageName=pck["ModelPackageArn"])

    logger.info("EndpointName= %s", endpoint_name)

    endpoints = sm.list_endpoints(NameContains=endpoint_name)['Endpoints']

    data_capture_config = DataCaptureConfig(
        enable_capture=True, sampling_percentage=100, destination_s3_uri=data_capture_s3_uri
    )

    logger.info("Data capture enabled")

    if len(endpoints) > 0:
        logger.info("Update current endpoint")

        model_statistics_s3_uri = model_description["ModelMetrics"]["ModelQuality"]["Statistics"]["S3Uri"]
        update_endpoint(sm, endpoint_name, data_capture_config, model_statistics_s3_uri)

    else:
        logger.info("Create endpoint")

        model_package_arn = model_description["ModelPackageArn"]
        create_endpoint(model_package_arn, sagemaker_session, instance_count, instance_type, endpoint_name,
                        data_capture_config)


def update_endpoint(sm, endpoint_name, data_capture_config, model_statistics_s3_uri):
    des_end_conf = sm.describe_endpoint_config(EndpointConfigName=endpoint_name)
    model_deployed_description = sm.describe_model(ModelName=des_end_conf["ProductionVariants"][0]["ModelName"])
    model_deployed_description = sm.describe_model_package(
        ModelPackageName=model_deployed_description["Containers"][0]["ModelPackageName"])

    new_model_metrics = load_json_from_s3(model_statistics_s3_uri)
    old_model_metrics = load_json_from_s3(
        model_deployed_description["ModelMetrics"]["ModelQuality"]["Statistics"]["S3Uri"])

    if new_model_metrics["binary_classification_metrics"]["accuracy"] > \
            old_model_metrics["binary_classification_metrics"]["accuracy"]:

        predictor = Predictor(endpoint_name=endpoint_name)
        predictor.update_endpoint(model_name=des_end_conf["ProductionVariants"][0]["ModelName"])
        predictor.update_data_capture_config(data_capture_config)
    else:
        logger.info(
            "Current endpoint is not updated because the new model have worse quality than current deployd model"
        )


def create_endpoint(model_package_arn, sagemaker_session,
                    instance_count, instance_type, endpoint_name,
                    data_capture_config):
    role = get_execution_role()
    model = ModelPackage(
        role=role, model_package_arn=model_package_arn, sagemaker_session=sagemaker_session
    )

    model.deploy(initial_instance_count=instance_count, instance_type=instance_type, endpoint_name=endpoint_name,
                 data_capture_config=data_capture_config)



import json
import boto3
import logging
from pathlib import Path
from omegaconf import OmegaConf
from botocore.client import BaseClient
from typing import Dict, List


def get_datetime_str(arg_dt):
    return arg_dt.strftime('%Y-%m-%d-%H-%M-%S')


def param_dict_to_nv_list(arg_dict):
    return [
        {'Name': k, 'Value': v}
        for k, v in arg_dict.items()
    ]


def convert_param_dict_to_key_value_list(arg_dict: Dict[str, str]) -> List[Dict[str, str]]:
    """
    Convert python dict to Sagemaker SDK resource tags structure
    where dict key corresponds to "Key", dict value corresponds to "Value".
    :param arg_dict: key-value need to convert to AWS resource tags structure
    :return: list of tags in the following format: [ { "Key": "...", "Value": "..." }, ... ]
    """
    return [
        {'Key': k, 'Value': v}
        for k, v in arg_dict.items()
    ]


def normalize_key(key: str):
    return key.replace('_', '').lower()


def get_model_name(model_arn):
    return model_arn[model_arn.rindex('/') + 1:]


def get_job_name(job_arn):
    return job_arn[job_arn.rindex('/') + 1:]


def get_output_destination(sm, processing_job_arn, output_name):
    pj_output_dict = _list_to_dict(
        sm.describe_processing_job(
            ProcessingJobName=get_job_name(processing_job_arn)
        )['ProcessingOutputConfig']['Outputs'],
        'OutputName'
    )
    return pj_output_dict[output_name]['S3Output']['S3Uri']


def get_model_location(sm, model_name):
    model_desc = sm.describe_model(ModelName=model_name)
    return model_desc['PrimaryContainer']['ModelDataUrl']


def _list_to_dict(arg_list, dict_key_attr):
    return {
        o[dict_key_attr]: o
        for o in arg_list
    }


def ensure_min_length(arg_str, min_length):
    if len(arg_str) < min_length:
        return arg_str + '_' + ('0' * (min_length - len(arg_str)))
    else:
        return arg_str


def get_approved_package(sm: BaseClient, model_package_group_name: str):
    response = sm.list_model_packages(
        ModelApprovalStatus='Approved',
        ModelPackageGroupName=model_package_group_name,
        SortBy='CreationTime',
        SortOrder='Descending',
        MaxResults=1,
    )

    model_packages = response['ModelPackageSummaryList']
    if len(model_packages) == 0:
        error_message = (
            f"No approved ModelPackage found for ModelPackageGroup: {model_package_group_name}"
        )
        raise Exception(error_message)

    return model_packages[0]


def load_json_from_s3(s3_uri: str):
    s3_client = boto3.client('s3')
    bucket, key = s3_uri.replace("s3://", "").split("/", 1)
    s3_response_object = s3_client.get_object(Bucket=bucket, Key=key)

    return json.loads(s3_response_object['Body'].read().decode('utf-8'))


def get_configs(mod_pipe, pipeline_module, args):
    default_conf_path = Path(mod_pipe.__file__).parent / (pipeline_module + '.defaults.yml')
    default_conf = OmegaConf.load(default_conf_path)
    override_arg_conf = OmegaConf.from_dotlist(args)
    return OmegaConf.merge(default_conf, override_arg_conf)


def _normalize_pipeline_name(name: str) -> str:
    name_max_len = 82
    name_len = len(name)
    if name_len > name_max_len:
        logging.getLogger(__name__).info(
            'Provided pipeline name "%s" is too long (%d symbols, max allowed: 82). It will be truncated.',
            name,
            name_len
        )
        name = name[:name_max_len]
    return name



import setuptools

setuptools.setup(
  name = 'mlops_sm',  
  version=open("version", "r").read().splitlines()[0],
  packages = ['mlops_sm'],   
  license='MIT',       
  description = 'Utility package for MlOps project',   
  author = 'Provectus Team',                   
  author_email = 'abelov@provectus.com',     
  keywords = ['MLOps', 'Utility package'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'boto3',
          'sagemaker',
          'setuptools',
          'datetime',
          'omegaconf', 
          'importlib',
          'pathlib',
          'botocore'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
